@@ -30,6 +30,7 @@ import Data.Map.Strict ( Map )
 import qualified Data.Map.Strict as Map
 import qualified Data.HashSet as Set
 import Control.Monad.State.Strict
+import Control.Monad.Identity
 import Data.SRTree.Recursion (cataM)
 import Algorithm.EqSat.Info
 import qualified Data.IntSet as IntSet
@@ -263,9 +264,21 @@ createDB = do modify' $ over (eDB . patDB) (const Map.empty)
               gets (_patDB . _eDB)
 {-# INLINE createDB #-}
 
+createDBBest :: Monad m => EGraphST m DB
+createDBBest = do modify' $ over (eDB . patDB) (const Map.empty)
+                  ecls <- gets (Prelude.map (\(eId, ec) -> (_best (_info ec), eId)) . IntMap.toList . _eClass)
+                  mapM_ (uncurry addToDB) ecls
+                  gets (_patDB . _eDB)
+
 -- | `addToDB` adds an e-node and e-class id to the database
 addToDB :: Monad m => ENode -> EClassId -> EGraphST m () -- State DB ()
-addToDB enode eid = do
+addToDB enode' eid = do
+  eid' <- canonical eid
+  isConst <- gets (_consts . _info . (IntMap.! eid') . _eClass)
+  let enode = case isConst of
+                ConstVal x -> Const x
+                ParamIx  x -> Param x
+                _          -> enode'
   let ids = eid : childrenOf enode -- we will add the e-class id and the children ids
       op  = getOperator enode    -- changes Bin op l r to Bin op () () so `op` as a single entry in the DB
   trie <- gets ((Map.!? op) . _patDB . _eDB)       -- gets the entry for op, if it exists
@@ -384,6 +397,11 @@ fromTrees :: Monad m => CostFun -> [Fix SRTree] -> EGraphST m [EClassId]
 fromTrees costFun = foldM (\rs t -> do eid <- fromTree costFun t; pure (eid:rs)) []
 {-# INLINE fromTrees #-}
 
+countParamsEg :: EGraph -> EClassId -> Int
+countParamsEg eg rt = countParams . runIdentity $ getBestExpr rt `evalStateT` eg
+countParamsUniqEg :: EGraph -> EClassId -> Int
+countParamsUniqEg eg rt = countParamsUniq . runIdentity $ getBestExpr rt `evalStateT` eg
+
 
 -- | gets the best expression given the default cost function
 getBestExpr :: Monad m => EClassId -> EGraphST m (Fix SRTree)
@@ -454,14 +472,60 @@ getAllExpressionsFrom eId' = do
         pure (t:ts)
 {-# INLINE getAllExpressionsFrom #-}
 
+getNExpressionsFrom :: Monad m => Int -> EClassId -> EGraphST m [Fix SRTree]
+getNExpressionsFrom n eId' = getNExpressionsFrom' n 15 eId' 
+
+getNExpressionsFrom' :: Monad m => Int -> Int -> EClassId -> EGraphST m [Fix SRTree]
+getNExpressionsFrom' _ 0 _ = pure []
+getNExpressionsFrom' n d eId' = do
+  eId <- canonical eId'
+  nodes <- gets (map decodeEnode . Set.toList . _eNodes . (IntMap.! eId) . _eClass)
+  (concat <$> go n d nodes)
+  where
+    isTerm (Var _) = True
+    isTerm (Const _) = True
+    isTerm (Param _) = True
+    isTerm _ = False
+    toTree (Var ix) = Fix $ Var ix
+    toTree (Const x) = Fix $ Const x
+    toTree (Param ix) = Fix $ Param ix
+    toTree _ = undefined
+
+    go n' _ []     = pure []
+    go n' 0 ts     = pure []
+    go n' d (node:ns) = do
+        tt <- Prelude.map Fix <$> case node of
+                Bin op l r -> do l' <- getNExpressionsFrom' n' (d-1) l
+                                 r' <- getNExpressionsFrom' n' (d-1) r
+                                 pure $ Prelude.take n [Bin op li ri | li <- l', ri <- r']
+                Uni f t    -> Prelude.map (Uni f) <$> getNExpressionsFrom' n' (d-1) t
+                Var ix     -> pure [Var ix]
+                Const x    -> pure [Const x]
+                Param ix   -> pure [Param ix]
+        let n'' = n' - length tt
+        if n'' <= 0
+          then pure [tt]
+          else do ts <- go n'' (d-1) ns
+                  pure (tt:ts)
+
 getAllChildEClasses :: Monad m => EClassId -> EGraphST m [EClassId]
 getAllChildEClasses eId' = do
   eId <- canonical eId'
-  nodes <- gets (map decodeEnode . Set.toList . _eNodes . (IntMap.! eId) . _eClass)
-  nub <$> go eId
+  IntSet.toList <$> go [eId] IntSet.empty
 
   where
-    go :: Monad m => Int -> EGraphST m [Int]
+    hasNoTerminal :: [ENode] -> Bool
+    hasNoTerminal = all (not . null . childrenOf) 
+    getNodes :: Monad m => EClassId -> EGraphST m [ENode]
+    getNodes n = gets (map decodeEnode . Set.toList . _eNodes . (IntMap.! n) . _eClass)
+
+    go :: Monad m => [Int] -> IntSet.IntSet -> EGraphST m IntSet.IntSet
+    go [] visited = pure visited
+    go queue visited = do 
+        nodes <- concatMap childrenOf . concat . filter hasNoTerminal <$> mapM getNodes queue
+        eids <- filter (\e -> e `IntSet.notMember` visited) <$> (mapM canonical nodes)
+        go eids (visited `IntSet.union` IntSet.fromList queue)
+            {-
     go n = do nodes <- gets (map decodeEnode . Set.toList . _eNodes . (IntMap.! n) . _eClass)
               let hasTerminal = any (null . childrenOf) nodes
               eids <- mapM canonical $ concatMap childrenOf nodes
@@ -469,8 +533,24 @@ getAllChildEClasses eId' = do
                 then pure [n]
                 else do eids' <- mapM go eids
                         pure ((n : eids) <> concat eids')
+                        -}
 {-# INLINE getAllChildEClasses #-}
 
+getAllChildBestEClasses :: Monad m => EClassId -> EGraphST m [EClassId]
+getAllChildBestEClasses eId' = do
+  eId <- canonical eId'
+  nub <$> go eId
+
+  where
+    go :: Monad m => Int -> EGraphST m [Int]
+    go n = do node <- gets (_best . _info . (IntMap.! n) . _eClass)
+              let hasTerminal = (null . childrenOf) node
+              eids <- mapM canonical $ childrenOf node
+              if hasTerminal
+                then pure [n]
+                else do eids' <- mapM go eids
+                        pure ((n : eids) <> concat eids')
+
 -- | returns a random expression rooted at e-class `eId`
 getRndExpressionFrom :: EClassId -> EGraphST (State StdGen) (Fix SRTree)
 getRndExpressionFrom eId' = do
